from typing import Any, Sequence, Mapping, Optional
import os
import psycopg

from .base_code_block import BaseCodeBlock


class SQLCodeBlock(BaseCodeBlock):
    """
    The SQLCodeBlock provides basic access to the Bismuth SQL service.
    """
    conn: psycopg.Connection

    def __init__(self, pg_uri: Optional[str] = None):
        if pg_uri:
            self.conn = psycopg.connect(pg_uri)
        else:
            self.conn = psycopg.connect(
                host="169.254.169.254",
                user="bismuth",
                password=os.environ['BISMUTH_AUTH'],
                dbname="bismuth",
            )

    def execute(self, sql: str, params: Sequence[Any]|Mapping[str, Any] = {}):
        """
        Run the given SQL, returning no results.
        """
        with self.conn.cursor() as cur:
            with self.conn.transaction():
                cur.execute(sql, params)

    def fetchone(self, sql: str, params: Sequence[Any]|Mapping[str, Any] = {}) -> Optional[tuple[Any, ...]]:
        """
        Run the given SQL, returning a single row.
        """
        with self.conn.cursor() as cur:
            with self.conn.transaction():
                cur.execute(sql, params)
                return cur.fetchone()

    def fetchall(self, sql: str, params: Sequence[Any]|Mapping[str, Any] = {}) -> list[tuple[Any, ...]]:
        """
        Run the given SQL, returning all rows.
        """
        with self.conn.cursor() as cur:
            with self.conn.transaction():
                cur.execute(sql, params)
                return cur.fetchall()

    def getconn(self) -> psycopg.Connection:
        """
        Get the underlying psycopg connection.
        """
        return self.conn



class BaseCodeBlock():
    """
    This class includes base methods and attributes for all other CodeBlocks
    """
    def __init__(self):
        """
        Initializes the base code block.
        """
        pass



import os
import requests
from flask import request
from http import HTTPStatus
from typing import Optional
from urllib.parse import urljoin

from .base_code_block import BaseCodeBlock


class ConfigurationCodeBlock(BaseCodeBlock):
    """
    The ConfigurationCodeBlock provides access to runtime configuration set outside the code.
    Instances of this class expose configuration to the application via the get method,
    which must be passed a string literal, NOT a variable.
    All instances of this class must be called `config`.
    """

    def __init__(self, api_url="http://169.254.169.254:9000/secrets/v1/"):
        """
        Initialize the configuration store.
        """
        if 'BISMUTH_AUTH' not in os.environ:
            print("Missing BISMUTH_AUTH token in environment. Configuration will not be available.")
            return
        self.auth = {"Authorization": "Bearer " + os.environ['BISMUTH_AUTH']}
        self.api_url = api_url

    def _headers(self):
        hdrs = self.auth.copy()
        try:
            for tracehdr in ["traceparent", "tracestate"]:
                if tracehdr in request.headers:
                    hdrs[tracehdr] = request.headers[tracehdr]
        except RuntimeError:
            pass
        return hdrs

    def get(self, key) -> Optional[str]:
        """
        Get a config value. Raises an exception if the key does not exist in the managed configuration store.
        """
        if 'BISMUTH_AUTH' not in os.environ:
            raise Exception("Missing BISMUTH_AUTH token in environment.")
        resp = requests.get(urljoin(self.api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise AttributeError(f"Configuration key {key} not set")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")
        return resp.text



from .base_code_block import BaseCodeBlock
from typing import ParamSpec, TypeVar, Generic, Callable


P = ParamSpec('P')
R = TypeVar('R')


class FunctionCodeBlock(BaseCodeBlock, Generic[P, R]):
    """
    Extends BaseCodeBlock. Run a python function with some requirements.
    """

    # Flag for if this block makes any network requests.
    network_enabled: bool
    # Function that will be run by `execute()`.
    func: Callable[P, R]

    def __init__(
        self, func: Callable[P, R], network_enabled: bool = False, *args, **kwargs
    ):
        """
        Initializes the FunctionCodeBlock with the function callable.
        """
        super().__init__(*args, **kwargs)
        self.network_enabled = network_enabled
        self.func = func

    def exec(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """
        Run a subset of arbitrary python functions as defined by Bismuth provided args and kwargs.
        """
        return self.func(*args, **kwargs)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self.exec(*args, **kwargs)



from .api_code_block import *
from .api_code_block_with_storage import *
from .auth_code_block import *
from .base_code_block import *
from .configuration_code_block import *
from .data_storage_code_block import *
from .function_code_block import *
from .persistent_data_storage_code_block import *
from .sql_code_block import *


from abc import ABC, abstractmethod
from typing import Optional
from functools import wraps
from flask import request, jsonify
from .base_code_block import BaseCodeBlock


class AuthService(ABC):
    @abstractmethod
    def validate_token(self, token) -> bool:
        """
        Validate the given token is authenticated
        """


class AuthCodeBlock(BaseCodeBlock):
    """
    Extends BaseCodeBlock. This class provides authentication functionalities for API endpoints.
    """
    # Holds the instance of the authentication service used for token validation.
    auth_service: AuthService

    def __init__(self, auth_service: Optional[AuthService] = None):
        """
        Initializes the AuthCodeBlock instance with a given authentication service.
        """
        if auth_service is None:
            self.auth_service = MockAuthService()
        else:
            self.auth_service = auth_service

    def token_required(self, f):
        """
        A decorator method to enforce token validation on API endpoints.
        It checks for a valid authentication token in the request header.
        """
        @wraps(f)
        def decorated(*args, **kwargs):
            token = None

            if "Authorization" in request.headers:
                auth_header = request.headers["Authorization"]
                if auth_header.startswith("Bearer "):
                    token = auth_header.split(" ")[1]

            if not token:
                return jsonify({"message": "Token is missing!"}), 401

            if not self.auth_service.validate_token(token):
                return jsonify({"message": "Token is invalid!"}), 401

            return f(*args, **kwargs)

        return decorated


# Example of a mocked authentication service
class MockAuthService:
    @staticmethod
    def validate_token(token):
        # Replace this with real validation logic
        return token == "valid_token"



from typing import Optional, Any, Callable, Concatenate
from flask import Flask, Request, request
from flask_restx import Api, Resource
from .auth_code_block import AuthCodeBlock
from .base_code_block import BaseCodeBlock
from .configuration_code_block import ConfigurationCodeBlock


class APICodeBlock(BaseCodeBlock):
    """
    Extends BaseCodeBlock, this class includes methods and attributes for API code blocks.
    """

    # The Flask application instance.
    app: Flask
    api: Api
    # The ConfigurationCodeBlock instance for this API.
    config: ConfigurationCodeBlock
    # The AuthCodeBlock instance for this API.
    auth_code_block: Optional[AuthCodeBlock]

    def __init__(
        self,
        title="API",
        version="1.0",
        description="A simple API",
        config: Optional[ConfigurationCodeBlock] = None,
        auth_code_block: Optional[AuthCodeBlock] = None,
        *args,
        **kwargs
    ):
        """
        Initializes the APICodeBlock class with optional title, version, description, configuration code block, and authentication code block.
        `add_route` should be called in overrides of this to register routes for the API.
        """
        super().__init__(*args, **kwargs)
        self.app = Flask(__name__)

        @self.app.route("/healthz")
        def healthz():
            return "ok"

        if config is None:
            config = ConfigurationCodeBlock()

        self.config = config
        # Hack: set prefix so that this doesn't register over the root route
        self.api = Api(
            self.app,
            version=version,
            title=title,
            description=description,
            doc="/doc",
            prefix="/tmp",
        )
        # And clear it
        self.api.prefix = ""
        self.auth_code_block = auth_code_block

    def _auth_handler(
        self,
        method: str,
        request: Request,
        handlers: dict[str, Callable[Concatenate[Request, ...], Any]],
        require_auth: list[str],
        **kwargs: Any
    ) -> Any:
        cb = handlers[method]

        if (
            method.upper() in map(lambda x: x.upper(), require_auth)
            and self.auth_code_block is not None
        ):
            return self.auth_code_block.token_required(cb)(request, **kwargs)
        else:
            return cb(request, **kwargs)

    def add_route(
        self,
        route: str,
        handlers: dict[str, Callable[Concatenate[Request, ...], Any]],
        require_auth: list[str] = ["POST", "DELETE", "PUT"],
    ):
        """
        Adds a route to the API.
        Handlers receive the request from flask plus query parameters as kwargs.
        """
        handlers = {k.upper(): v for k, v in handlers.items()}
        auth_handler = self._auth_handler

        class DynamicResource(Resource):
            def get(self):
                kwargs = request.args.to_dict()
                if "GET" in handlers:
                    return auth_handler(
                        "GET", request, handlers, require_auth, **kwargs
                    )
                else:
                    self.api.abort(404)

            def post(self):
                kwargs = request.get_json(silent=True) or {}
                if "POST" in handlers:
                    return auth_handler(
                        "POST", request, handlers, require_auth, **kwargs
                    )
                else:
                    self.api.abort(404)

            def put(self):
                kwargs = request.get_json(silent=True) or {}
                if "PUT" in handlers:
                    return auth_handler(
                        "PUT", request, handlers, require_auth, **kwargs
                    )
                else:
                    self.api.abort(404)

            def delete(self):
                kwargs = request.args.to_dict()

                if "DELETE" in handlers:
                    return auth_handler(
                        "DELETE", request, handlers, require_auth, **kwargs
                    )
                else:
                    self.api.abort(404)

        self.api.add_resource(DynamicResource, route)

    def __call__(self, *args):
        return self.app(*args)

    def run(self, host="0.0.0.0", port=5000, debug=False):
        """
        Starts the API on the given port and host.
        """

        self.app.run(host=host, port=port, debug=debug)



import os
import pytest
from flask import Request
from .api_code_block import APICodeBlock
from .function_code_block import FunctionCodeBlock


# Fixture to setup the APICodeBlock with mocked routes
@pytest.fixture
def api_block():
    os.environ["BISMUTH_AUTH"] = "TEST_AUTH"
    api_block = APICodeBlock()
    api_block.app.testing = True
    return api_block


def test_request_passed_through_auth_callback(api_block):
    def request_passed(request: Request):
        if type(request) != Request:
            return 500
        else:
            return 200

    api_block.add_route("/request_passed", {"GET": request_passed})

    with api_block.app.test_client() as client:
        response = client.get("/request_passed")
        assert response.status_code == 200


def test_add_route(api_block):
    def func(request, **kwargs):
        return {"message": "mock response"}

    api_block.add_route("/mock", {"get": FunctionCodeBlock(func)})

    with api_block.app.test_client() as client:
        response = client.get("/mock")
        assert response.status_code == 200
        assert response.json == {"message": "mock response"}


def test_add_route_bare_func(api_block):
    def func(request, **kwargs):
        return {"message": "mock response"}

    api_block.add_route("/mock", {"get": func})

    with api_block.app.test_client() as client:
        response = client.get("/mock")
        assert response.status_code == 200
        assert response.json == {"message": "mock response"}


def test_add_root_route(api_block):
    def func(request, **kwargs):
        return {"message": "mock response"}

    api_block.add_route("/", {"get": FunctionCodeBlock(func)})

    with api_block.app.test_client() as client:
        response = client.get("/")
        assert response.status_code == 200
        assert response.json == {"message": "mock response"}


def test_json_kwargs(api_block):
    def func(request, **kwargs):
        return {"json": kwargs}

    api_block.add_route("/", {"post": FunctionCodeBlock(func)})

    with api_block.app.test_client() as client:
        response = client.post("/", json={"key": "value"})
        assert response.status_code == 200
        assert response.json == {"json": {"key": "value"}}


def test_non_json_request(api_block):
    def func(request, **kwargs):
        return {"data": request.get_data(as_text=True)}

    api_block.add_route("/", {"post": FunctionCodeBlock(func)})

    with api_block.app.test_client() as client:
        response = client.post("/", data={"form": "value"})
        assert response.status_code == 200
        assert response.json == {"data": "form=value"}



import json
import os
import pathlib
import requests
import shutil
from flask import request
from http import HTTPStatus
from typing import Optional, Dict, Any
from urllib.parse import urljoin

from .data_storage_code_block import DataStorageCodeBlock


class PersistentDataStorageCodeBlock(DataStorageCodeBlock):
    """
    This class makes data storage operations persistent using Bismuth's blob storage service,
    allowing JSON-encodable objects to be persisted.
    When run locally, it stores data via files in a temporary directory.
    """

    def __init__(self, api_url="http://169.254.169.254:9000/blob/v1/"):
        """
        Initialize the datastore.
        """
        if 'BISMUTH_AUTH' in os.environ:
            self._impl = HostedPersistentDataStorageCodeBlock(os.environ['BISMUTH_AUTH'], api_url)
        else:
            self._impl = LocalPersistentDataStorageCodeBlock()

    def create(self, key, value) -> None:
        """Create a new item in the datastore."""
        return self._impl.create(key, value)

    def retrieve(self, key) -> Optional[Any]:
        """Retrieve an item from the datastore."""
        return self._impl.retrieve(key)

    def update(self, key, value) -> None:
        """Update an existing item in the datastore."""
        return self._impl.update(key, value)

    def delete(self, key) -> None:
        """Delete an item from the datastore."""
        return self._impl.delete(key)

    def list_all(self) -> Dict[str, Any]:
        """List all items in the datastore."""
        return self._impl.list_all()


class HostedPersistentDataStorageCodeBlock(DataStorageCodeBlock):
    # A dictionary of HTTP headers used to authenticate to the storage backend.
    _auth: Dict[str, str]
    # The URL of the storage backend.
    _api_url: str

    def __init__(self, auth: str, api_url="http://169.254.169.254:9000/blob/v1/"):
        self._auth = {"Authorization": "Bearer " + auth}
        self._api_url = api_url

    def _headers(self):
        hdrs = self._auth.copy()
        try:
            for tracehdr in ["traceparent", "tracestate"]:
                if tracehdr in request.headers:
                    hdrs[tracehdr] = request.headers[tracehdr]
        except RuntimeError:
            pass
        return hdrs

    def create(self, key, value) -> None:
        resp = requests.post(urljoin(self._api_url, key), data=json.dumps(value), headers=self._headers())
        if resp.status_code == HTTPStatus.CONFLICT:
            raise ValueError("Key already exists.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def retrieve(self, key) -> Optional[Any]:
        resp = requests.get(urljoin(self._api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            return None
        elif not resp.ok:
            raise Exception(f"Server error {resp}")
        return resp.json()

    def update(self, key, value) -> None:
        resp = requests.put(urljoin(self._api_url, key), data=json.dumps(value), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise ValueError("Key does not exist.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def delete(self, key) -> None:
        resp = requests.delete(urljoin(self._api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise ValueError("Key does not exist.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def list_all(self) -> Dict[str, Any]:
        resp = requests.get(self._api_url, headers=self._headers())
        if not resp.ok:
            raise Exception(f"Server error {resp}")
        return dict((k, json.loads(bytes(v))) for k, v in resp.json().items())


class LocalPersistentDataStorageCodeBlock(DataStorageCodeBlock):
    _dir = pathlib.Path("/tmp/bismuth_persistent_storage/")

    def create(self, key, value) -> None:
        path = self._dir / key
        if path.exists():
            raise ValueError("Key already exists.")
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._dir / key, 'w') as f:
            json.dump(value, f)

    def retrieve(self, key) -> Optional[Any]:
        path = self._dir / key
        if not path.exists():
            return None
        with open(path, 'r') as f:
            return json.load(f)

    def update(self, key, value) -> None:
        path = self._dir / key
        if not path.exists():
            raise ValueError("Key does not exist.")
        with open(self._dir / key, 'w') as f:
            json.dump(value, f)

    def delete(self, key) -> None:
        path = self._dir / key
        if not path.exists():
            raise ValueError("Key does not exist.")
        path.unlink()

    def list_all(self) -> Dict[str, Any]:
        out = {}
        for fn in self._dir.glob('**/*'):
            with open(fn, 'r') as f:
                out[str(fn.relative_to(self._dir))] = json.load(f)
        return out

    def clear(self) -> None:
        if self._dir.exists():
            shutil.rmtree(self._dir)